
###########################
#parseToNodes
###########################

require(codetools)

flowControlFunctions <- c("if","while","repeat","for",
	#rest is special
	"function"
	)
operatorFunctions <- c(
	"+", "-", "*", "/", "%", "^",
	">", ">=", "<", "<=", "==", "!=",
	"!", "&", "|",
	"~",
	"->", "<-",
	"$",
	":",
	"%%", "%*%", "%/%", "%in%", "%o%", "%x%",
	"(", "{",
	"[", "[[")
	
deparseTreeExp <- function(e){
	s <- NULL
	d <- deparse(e, width.cutoff = 400)
	if(length(d) > 1){
		for(x in d){
			s <- paste(s, x, "\n" ,sep="")
		}
		s <- substring(s, 1, nchar(s, type="chars")-1)
		return(s)
	}
	return(d)
}

parseToNodes <- function(text = NULL, file = "", keep.source=getOption("keep.source")){
	if (is.null(text) & (is.null(file) || nchar(file) == 0)) {
		return(list())
	}
	
	#save old keep.source option
	old.keep.source <- getOption("keep.source")
	#restore options on exit
	on.exit(options(keep.source=old.keep.source))
	#change keep.source option
	options(keep.source=keep.source)
	
	#"parse" uses keep.souce option
	codes <- parse(text = text, file = file)
	
	
	lines <- list()
	
	if (length(codes) > 0) {
		for(i in 1:length(codes)){
			e <- codes[[i]]
			flist <- retrieveFunction(e)
			lines[[i]] <- flist
		}
	}
	
	if(!keep.source){
		return(lines)
	}
	
	#if keep.source, add non code lines
	lines2 <- list()
	ref <- attr(codes, "srcref", exact=TRUE)
	wref <- attr(codes, "wholeSrcref", exact=TRUE)
	src <- attr(codes, "srcfile", exact=TRUE)
	if(!is.null(ref) && !is.null(wref) && !is.null(src)){
		codeIndex <- 1
		linesIndex <- 1
		lineNum <- wref[1]
		while(lineNum <= wref[3]){
			#skip last empty line
			if(lineNum == wref[3] && wref[4] == 0){
				break
			}
			if(codeIndex > length(ref) || lineNum < ref[[codeIndex]][1]){
				#comment or empty lines
				lines2[[linesIndex]] <- list(line=getSrcLines(src, lineNum, lineNum)[1])
				lines2[[linesIndex]]$srcref <- as.integer(c(lineNum-1, lineNum-1, 0, -1))
				linesIndex <- linesIndex + 1
				lineNum <- lineNum + 1
			}else{
				#code line
				while(codeIndex <= length(ref) && lineNum == ref[[codeIndex]][1]){
					lines2[[linesIndex]] <- lines[[codeIndex]]
					#srcLines <- getSrcLines(src, ref[[codeIndex]][1], ref[[codeIndex]][3])
					##cut line head
					#if(ref[[codeIndex]][5] != 0){
					#	srcLines[1] <- substring(srcLines[1], ref[[codeIndex]][5])
					#}
					##cut line tail
					##if next code is in the same line
					#if(length(ref) > codeIndex && ref[[codeIndex+1]][1]==lineNum){
					#	srcLines[length(srcLines)] <- substring(srcLines[length(srcLines)], 1, ref[[codeIndex]][6])
					#}
					
					#srcLines
					srcLines <- as.character(ref[[codeIndex]])
					#if next code is not in the same line, replace last line to whole line
					#to include line tail comment
					if (!(length(ref) > codeIndex && ref[[codeIndex+1]][1]==ref[[codeIndex]][3])){
						#lastLine <- getSrcLines(src, ref[[codeIndex]][3], ref[[codeIndex]][3])[1]
						#srcLines[length(srcLines)] <- lastLine
						refTemp <- ref[[codeIndex]]
						refTemp[c(4,6)] <- .Machine$integer.max
						srcLines <- as.character(refTemp)
					}
					
					#joint lines srcLines to srcLine
					srcLine <- ""
					for(x in srcLines){
						srcLine <- paste(srcLine, x, "\n", sep="")
					}
					srcLine <- substring(srcLine, 1, nchar(srcLine, type="chars")-1)
					
					lines2[[linesIndex]]$srcline <- srcLine
					lines2[[linesIndex]]$srcref <- ref[[codeIndex]][c(1,3,5,6)]-as.integer(1)
					
					lineNum <- ref[[codeIndex]][3]
					codeIndex <- codeIndex + 1
					linesIndex <- linesIndex + 1
				}
				lineNum <- lineNum + 1
			}
		}
	
		return(lines2)
	}else{
		return(lines)
	}
}


retrieveFunction <- function(e){
	env <- new.env()
	
	env$line = deparseTreeExp(e)
	
	functionCollect <- function(v, e){
		assign(v, e, envir = env)
	}
	w <- functionWalker(functionCollect)
	w$top = TRUE
	
	walkCode(e, w)
	
	as.list(env)
}


functionHandler <- function(v, w){
	#if(w$top){
		switch(v, "="=,"<-"=assignHandler)
	#}
}

assignHandler <- function(e, w) {
	w$collect("symbol", deparseTreeExp(e[[2]]))
	w$top = FALSE
	walkCode(e[[3]], w)
}

functionCall <- function(e, w){
	w$collect("sentence", deparseTreeExp(e))
	fun <- deparseTreeExp(e[[1]])
	if(fun %in% flowControlFunctions){
		w$collect("control", TRUE)
	}else{
		if(fun %in% operatorFunctions){
			#do nothing
		}else{
			w$collect("func", fun)
			args <- as.list(e[-1])
			args <- lapply(args, function(x){x <- deparseTreeExp(x)})
			w$collect("args", args)
		}
	}
}

functionLeaf <- function(e, w){
	w$collect("sentence", deparseTreeExp(e))
}

functionWalker <- function(collect){
	w <- makeCodeWalker(handler = functionHandler, call = functionCall, leaf = functionLeaf)
	w$collect = collect
	w
}

###########################
#r -> java function
###########################
.rflow.pager <- function(file, header, title, delete.file) {
  invisible(.jnew("com/ef_prime/rflow/tool/SimplePager",as.character(file), as.character(header), as.character(title), as.logical(delete.file)))
}
#options(pager=.rflow.pager)

###########################
#Rflow function
###########################

#RflowGD
#In JavaGD, we can't get the device number until plot.new. So always do plot.new.
RflowGD <- function(width = NULL, height = NULL, ps=NULL, ...){
	if ( is.null(width) || is.null(height) ){
		size = get(".RFLOW_GD_SIZE", envir = parent.frame())
		width = size[1]
		height = size[2]
	}
	if ( is.null(ps) ) {
		ps = get(".RFLOW_GD_POINT_SIZE", envir = parent.frame())
	}
	JavaGD::JavaGD(width=width, height=height, ps=ps, ...)
	par(bg = "transparent")
	plot.new()
}


#clear all objects of envir
.clearObjects <- function(envir = .GlobalEnv)
{
	rm(list = ls(envir = envir, all.names = TRUE), envir = envir)
}

#new graphics device
.newGraphics <- function(...)
{
	do.call(getOption("device"), list(...))
}

#set to next active device
#.rflowSetNextDevice <- function() {
#	a <- dev.list()
#	a <- a[a != dev.cur()]
#	if (length(a) == 0) {
#		return()
#	}
#
#	a <- max(a)
#	dev.set(a[1])
#}

#get function arguments
.rflowGetFormals <- function (fun)
{
	if (is.character(fun)) {
		fun <- get(fun, pos=1)
	}
	if (is.primitive(fun)) {
		formals <- formals(args(fun))
	} else {
		formals <- formals(fun)
	}
	for(i in 1:length(formals))
	{
		formals[[i]] <- deparse(formals[[i]])
	}
	return(formals)
}

#get function's local variables
.rflowFindLocals <- function (fun, merge = FALSE) 
{
    vars <- new.env(TRUE)
    funs <- new.env(TRUE)
    enter <- function(type, v, e, w) if (type == "function") 
        assign(v, TRUE, funs)
    else assign(v, TRUE, vars)
    collectUsage(fun, enterLocal = enter)
    fnames <- ls(funs, all.names = TRUE)
    vnames <- ls(vars, all.names = TRUE)
    if (merge) 
        sort(unique(c(vnames, fnames)))
    else list(functions = fnames, variables = vnames)
}

#do expr in envir
.rflowDoInLocal <- function(expr, envir, envchar, ...) 
{
	#if(is.null(envir)) {
	#  envir <- new.env(..., parent=.GlobalEnv)
	#  eval(expr, envir=envir)
	#  return(envir)
	#}
	
	if(!exists(envchar)){
		env <- new.env(parent=.GlobalEnv)
		assign(envchar, env, envir = .GlobalEnv)
		eval(expr, envir)
	}
	else{
		eval(expr, envir)
	}
}

#get names
# x is list of characters
#.rflowGetNames <- function(x) {
#	getname <- function(objName) {
#		if (exists(objName, where=1)) {
#			return(names(get(objName, pos=1)))
#		} else {
#			return(NULL)
#		}
#	}
#	return(lapply(x, getname))
#}

#command support
.rflowSupportCommand <- function(x, envir = NULL)
{
	ret <- NULL
	packages <- NULL
	
	#get environment names
	#remove preceding 'package:'
	searchNames <- search()
	searchNum <- length(searchNames)

	
	if((a <- regexpr('\\$$', x)) > 1){
		#ex 'iris$' 'Boston$'
		
		aList <- substr(x, 1, a-1)
		for(i in 1:searchNum) {
			if(exists(aList, where = i, inherits = FALSE)) {
				listNames <- names(get(aList, pos = i, inherits = FALSE))
				ret <- c(ret, listNames)
				packages <- c(packages, rep(i, length(listNames)))
			}
		}
	}else if((a <- regexpr('\\$', x)) > 1){
		#ex 'iris$Sp' 'Boston$c'
		
		aList <- substr(x, 1, a-1)
		y <- substr(x, a+1, nchar(x))
		pat <- paste('^', y, sep="")
		#escape dot
		pat <- gsub("\\.", "\\\\.", pat)
		
		for(i in 1:searchNum) {
			if(exists(aList, where = i, inherits = FALSE)) {
				listNames <- names(get(aList, pos = i, inherits = FALSE))
				hitIndices <- grep(pat, listNames, ignore.case=TRUE)
				ret <- c(ret, listNames[hitIndices])
				packages <- c(packages, rep(i, length(hitIndices)))
			}
		}
	}else{
		#escape dot
		x <- gsub("\\.", "\\\\.", x)
		pat <- paste("^", x, ".*", sep="")
		if(is.null(envir)){
			for(i in 1:searchNum) {
				namez <- grep(pat, ls(pos=i, all.names=TRUE), ignore.case=TRUE, value=TRUE)
				ret <- c(ret, namez)
				packages <- c(packages, rep(i, length(namez)))
			}
		}else{
			namez <- grep(pat, ls(pos=envir, all.names=TRUE), ignore.case=TRUE, value=TRUE)
			ret <- c(ret, namez)
			packages <- c(packages, rep(i, length(namez)))
		}
	}
	
	if (is.null(ret)) {
		return(NULL)
	}else{
		return(list(word=ret, package=packages, search=searchNames))
	}
}


#get objects filtered by FILTERFUN
#FILTERFUN must return TRUE/FALSE
.rflowGetObjectsFiltered <- function(envir = .GlobalEnv, FILTERFUN)
{
	a <- unlist(eapply(envir, FILTERFUN))
	if(!is.null(a)){
		a <- sort(names(which(a)))
	}
	a
}

#get length or rownum of obj
.rflowGetLength <- function(obj)
{
	tip <- dim(obj)
	if(is.null(tip)){
		tip <- length(obj)
		#if(len == 1){
		#	tip <- NULL
		#}else{
		#	tip <- len
		#}
	}else{
		tip <- tip[1]
	}
	
	return(tip)
}

#get help rd object
#topic: topic name of help
#filter: character vector of Rtag name. NULL for no filter
.rflowGetRd <- function(topic){
	rd <- utils:::.getHelpFile(help(topic))
	rd
}

.rflowGetArgHelps <- function(topic){
	tools:::.Rd_get_argument_table(.rflowGetRd(topic))
}

#get help html
#topic: topic name of help
#dynamic: html links if TRUE
.rflowGetHtmlHelp <- function(topic, dynamic=FALSE, stylesheet=NULL){
	helpDir <- help(topic)
	rd <- utils:::.getHelpFile(helpDir)
	html <- character()
	htmlCon <- textConnection("html", open="w", local=TRUE)
	on.exit(close(htmlCon))
	if (is.null(stylesheet)) {
		stylesheet <- file.path(R.home("doc"), "html", "R.css")
	}
	tools::Rd2HTML(Rd=rd, out=htmlCon, package=dirname(dirname(helpDir)),
		dynamic=dynamic, stylesheet=stylesheet, outputEncoding=localeToCharset())
	html
}


.rflowGetObject <- function(parent = .GlobalEnv, obj=NULL, name, fullName){
	if (is.null(obj)){
		obj <- get(fullName, envir=parent)
	}
	#if (is.character(obj)) {
	#	obj <- get(obj, envir=parent)
	#}
	
	#check if obj is valid
	type <- NULL
	try(suppressWarnings(type <- typeof(obj)), silent=true)
	if(is.null(type)){
		return(NULL)
	}
	objInfo <- list()
	objInfo$type <- typeof(obj)
	objInfo$class <- class(obj)
	
	#flag
	flag <- 0
	if(is.data.frame(obj)){flag <- flag+1}
	if(is.matrix(obj)){flag <- flag+2}
	if(is.array(obj)){flag <- flag+2^2}
	objInfo$flag <- as.integer(flag)
	
	dimu <- dim(obj)
	tip <- ""
	if(is.null(dimu)){
		len <- length(obj)
		if(len == 1){
			tip <- substring(deparse(obj, width.cutoff = 100), 1, 100)
			dimu <- as.integer(1)
		}else{
			dimu <- as.integer(len)
		}
	}
	#else{
	#	tip <- paste("[", as.character(tip[1]), ",", as.character(tip[2]), "]", sep="")
	#}
	objInfo$dim <- dimu
	objInfo$tip <- tip
	if (missing(name) || is.null(name)) {
		isExpression <- FALSE
	} else {
		isExpression <- length(parse(text=name)[[1]]) > 1
	}
	objInfo$validname <- ifelse(isExpression | is.numeric(name), TRUE, (make.names(name)==name))
	objInfo
}
	
	
#get objects informations
#all.names: see ls
#limit: maximum of obj infos to be returned
.rflowGetObjects <- function(parent = .GlobalEnv, all.names = FALSE, limit = 200)
{
	objects <- list()
	#attr name
	overlimit <- "overlimit"
	
	getObjInfo <- function(name, obj){
		objInfo <- list()
		#check if obj is valid
		type <- NULL
		try(suppressWarnings(type <- typeof(obj)), silent=true)
		if(is.null(type)){
			return(objInfo)
		}
		objInfo$type <- typeof(obj)
		objInfo$class <- class(obj)
		
		#flag
		flag <- 0
		if(is.data.frame(obj)){flag <- flag+1}
		if(is.matrix(obj)){flag <- flag+2}
		if(is.array(obj)){flag <- flag+2^2}
		objInfo$flag <- as.integer(flag)
		
		dimu <- dim(obj)
		tip <- ""
		if(is.null(dimu)){
			len <- length(obj)
			if(len == 1){
				tip <- substring(deparse(obj, width.cutoff = 100), 1, 100)
				dimu <- as.integer(1)
			}else{
				dimu <- as.integer(len)
			}
		}
		#else{
		#	tip <- paste("[", as.character(tip[1]), ",", as.character(tip[2]), "]", sep="")
		#}
		objInfo$dim <- dimu
		objInfo$tip <- tip
		objInfo$validname <- ifelse(is.numeric(name), TRUE, (make.names(name)==name))
		objInfo
	}
	
	parentType <- NULL
	try(parentType <- typeof(parent), silent=TRUE)
	if(is.null(parentType)){
		return(objects)
	}
	if(parentType == "environment"){
		objs <- ls(parent, all.names = all.names)
		if(length(objs) > 0){
			len <- ifelse(limit > 0 && length(objs) > limit, limit, length(objs))
			objects <- vector("list", len)
			for(i in 1:len){
				objects[[i]] <- getObjInfo(objs[i], parent[[objs[i]]])
				names(objects)[i] <- objs[i]
			}
			if(length(objs) > limit){
				attr(objects, overlimit) <- length(objs)
			}
		}
	}else if(parentType == "list" || parentType == "pairlist"){
		namez <- names(parent)
		if(is.null(namez)){
			#no names
			if(length(parent) > 0){
				len <- ifelse(limit > 0 && length(parent) > limit, limit, length(parent))
				objects <- vector("list", len)
				for(i in 1:len){
					objects[[i]] <- getObjInfo(i, parent[[i]])
					#names(objects)[i] <- i
				}
				if(length(parent) > limit){
					attr(objects, overlimit) <- length(parent)
				}
			}
		}else{
			if(length(namez) > 0){
				len <- ifelse(limit > 0 && length(namez) > limit, limit, length(namez))
				objects <- vector("list", len)
				for(i in 1:len){
					#empty name is index
					x <- ifelse(namez[i] == "", i, namez[i])
					objects[[i]] <- getObjInfo(x, parent[[i]])
					names(objects)[i] <- namez[i]
				}
				if(length(namez) > limit){
					attr(objects, overlimit) <- length(namez)
				}
			}
		}
	}
	
	return(objects)
}

#check whether in browse or not
#if in browse, .browserObject is value of browserText(), otherwise NULL
#if in browse, .debugEnvironment is set
#if in browse, .rflowCurrentCall is set
.rflowCheckBrowser <- function(envir){
	tryCatch(assign(".rflowBrowserObject", browserText(), envir=envir), error=function(e){assign(".rflowBrowserObject", NULL, envir=envir)})
	brObj <- get(".rflowBrowserObject", envir=envir)
	if (!is.null(brObj)){
		index <- -1
		if(brObj==".rflowRecover"){
			index <- -1:-4
		}
		
		frames <- sys.frames()
		try(assign(".rflowDebugEnvironment", rev(frames)[index], envir=envir), silent=TRUE)
		
		calls <- sys.calls()
		try(assign(".rflowCurrentCall", lapply(rev(calls)[index], format), envir=envir), silent=TRUE)
		
	} else {
		try(assign(".rflowDebugEnvironment", list(globalenv())), silent=TRUE)
	}
	invisible()
}

#save error state
# .rflowErrorHandler <- function(){
# 	#rflowEnv <- as.environment(".RflowEnv")
# 	rflowEnv <- parent.frame()
# 	
# 	#set error object
# 	try(assign(".rflowErrorObject", TRUE, envir=rflowEnv), silent=TRUE)
# 	
# 	#browse if .rflowRecover is set
# 	recoverFlag <- get(".rflowRecover", envir=rflowEnv)
# 	which <- -1
# 	currentEnv <- get(".rflowDebugEnvironment", envir=rflowEnv)[[1]]
# 	if(!is.null(recoverFlag) && recoverFlag && !identical(sys.frame(which), currentEnv)){
# 		#eval(substitute(browser(text = callName, skipCalls = 3), list(callName = format(sys.call(which)))), envir = sys.frame(which))
# 		#eval(substitute(browser(text = callName, skipCalls = 3), list(callName = format(sys.call(which)))), envir = sys.frame(which))
# 		eval(substitute(browser(text = ".rflowRecover", skipCalls = 3)), envir = sys.frame(which))
# 	}
# 	
# 	invisible()
# }

#variable for error check
.rflowErrorObject <- NULL

#varibale for browser check
.rflowBrowserObject <- NULL

#variable for refering debug environment
.rflowDebugEnvironment <- list(globalenv())

#variable for refering current call
.rflowCurrentCall <- NULL

#variable for using recover
.rflowRecover <- FALSE

#variable for debug
.RFLOW_DEBUG <- FALSE

#variable for graphics device size
#7inch in 72dpi = 504px
#7inch in 96dpi = 672px
.RFLOW_GD_SIZE <- c(600, 600)
.RFLOW_GD_POINT_SIZE <- 14

#preview environment
.PREVIEW_ENV <- new.env()
